#include <windows.h>

#include <cstring>
#include <queue>

#include "common/GMLuaModule.h"

GMOD_MODULE( Init, Shutdown );

struct BeepInfo
{
	unsigned short frequency;
	DWORD duration;
};

std::queue<BeepInfo*> BeepQueue;

HANDLE BeeperThread = NULL;
bool BeeperActive = true;
bool BeepLock = false;

BeepInfo* CurrentBeep = NULL;

static void BeepHandler()
{
	while(BeeperActive)
	{
		if(!BeepQueue.empty())
		{
			while(BeepLock) Sleep(0);

			BeepLock = true;
			BeepInfo* beep = CurrentBeep = BeepQueue.front();
			BeepQueue.pop();
			BeepLock = false;

			if(beep->frequency == 0)
				Sleep(beep->duration);
			else
				Beep(beep->frequency, beep->duration);
			
			delete beep;
		}
		else
		{
			CurrentBeep = NULL;
			SuspendThread(BeeperThread);
		}
	}
}

/* __call function */
LUA_FUNCTION( BeepCall )
{
	unsigned short freq;
	DWORD dur;

	g_Lua->CheckType(2, GLua::TYPE_NUMBER);

	if(g_Lua->Top() == 2)
	{
		freq = 0; //meaning this entry is a pause
		dur = g_Lua->GetInteger(2);
	}
	else
	{
		freq = g_Lua->GetInteger(2);
		g_Lua->CheckType(3, GLua::TYPE_NUMBER);
		dur = g_Lua->GetInteger(3);
	}

	BeepInfo* beep = new BeepInfo();
	beep->duration = dur;
	beep->frequency = freq;

	while(BeepLock) Sleep(0);
	
	BeepLock = true;

	if(BeepQueue.empty())
		CurrentBeep = beep;

	BeepQueue.push(beep);
	BeepLock = false;

	ResumeThread(BeeperThread);
	return 0;
}

/*__index function*/
LUA_FUNCTION( BeepIndex )
{
	const char* key = g_Lua->GetString(2);
	
	if(strcmp(key, "Beeping") == 0)
		g_Lua->Push( ResumeThread(BeeperThread) == 0 );

	else if(strcmp(key, "Queue") == 0)
		g_Lua->Push( static_cast<float>(BeepQueue.size()) );

	else if(strcmp(key, "Frequency") == 0)
		g_Lua->Push(CurrentBeep? CurrentBeep->frequency : static_cast<float>(0));

	else if(strcmp(key, "Duration") == 0)
	g_Lua->Push(CurrentBeep? CurrentBeep->duration : static_cast<float>(0));

	else
		g_Lua->LuaError("Only \"Beeping\", \"Queue\", \"Frequency\" and \"Duration\" allowed\n");

	return 1;
}

int Init( void )
{
	g_Lua->NewTable();
	ILuaObject* meta = g_Lua->GetObject();
	
	g_Lua->Push(BeepCall);
	meta->SetMember("__call");

	g_Lua->Push(BeepIndex);
	meta->SetMember("__index");

	g_Lua->NewGlobalTable("Beep");
	ILuaObject* table = g_Lua->GetGlobal("Beep");
	g_Lua->SetMetaTable(table, meta);

	SAFE_UNREF(meta);
	SAFE_UNREF(table);

	BeeperThread = CreateThread(NULL, 0, (LPTHREAD_START_ROUTINE)&BeepHandler, NULL, 0, NULL);
	return 0;
}

int Shutdown( void )
{
	BeeperActive = false;
	ResumeThread(BeeperThread);
	return 0;
}

